import java.text.Normalizer;

/*******************************************************************************
 * Třída CaesarovaSifra ...
 *
 * @author    Luboš Pavlíček
 * @version   1.00.000,  19. 2. 2014
 */
public class VigenerovaSifra implements Sifra {

    //== Datové atributy (statické i instancí)======================================
    // proměnná definuje přípustné znaky po normalizaci otevřeném textu (a i v zašifrovaném textu)
    // používám písmena velké anglické abecedy.
    private static final char[] ABECEDA= {
            'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
            'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z' };

    // proměnná bude obsahovat posuny abecedy pro kazde pismeno hesla (A => 0, B => 1, C => 2, ...)
    private int[] posuny;

    //== Konstruktory a tovární metody =============================================

    /***************************************************************************
     * Konstruktor pro Vigenerovu šifru. Parametrem je heslo.
     * 
     * @param heslo heslo používané pro šifrování/dešifrování
     */
    public VigenerovaSifra(String heslo) {
        if ((heslo == null)  || (heslo.isEmpty())) {
            throw new IllegalArgumentException("heslo musí obsahovat aspoň jeden znak ");
        }
        String hesloNorm = normalizuj(heslo);
        if (hesloNorm.isEmpty()) {
            throw new IllegalArgumentException("v hesle chybí písmena či číslice ");
        }
        posuny = new int[hesloNorm.length()];
        for (int i=0; i < hesloNorm.length(); i++) {
            for (int j=0; j < ABECEDA.length; j++) {
                if (hesloNorm.charAt(i) == ABECEDA[j]) {
                    posuny[i] = j;
                    break;
                }
            }
        }
    }

    //== Nesoukromé metody (instancí i třídy) ===============================================
    /**
     * Metoda zašifruje vstupní text za pomocí caesarovy šifry. Pokud je na vstupu neznámý znak,
     * vznikne výjimka IllegalArgumentException.
     * 
     * @param otevrenyText   vstupní text k zašifrování
     * @return               zašifrovaný text
     */
    public String zasifruj(String otevrenyText) {
        if (otevrenyText == null) {
            throw new IllegalArgumentException("otevřený text musí obsahovat aspoň jeden znak ");
        }
        // nejdříve pomocí metody normalizuj zruším diakritiku, převedu na velká písmena,
        String normalText = normalizuj(otevrenyText);
        if (normalText.isEmpty()) {
            throw new IllegalArgumentException("v otevřeném textu chybí písmena/číslice ");
        }
        // pro sifrovyText použiji StringBuilder (rychlost)
        StringBuilder sifrovyText = new StringBuilder(normalText.length());
        for (int i=0; i< normalText.length(); i++) {
            sifrovyText.append(zasifrujZnak(normalText.charAt(i), i));
        }
        return sifrovyText.toString();
    }

    /**
     * Metoda dešifruje vstupní text za pomocí caesarovy šifry. Pokud je na vstupu neznámý znak,
     * vznikne výjimka IllegalArgumentException.
     * 
     * @param sifrovyText   vstupní text k dešifrování
     * @return              otevřený text
     */
    public String desifruj(String sifrovyText) {
        if (sifrovyText == null || sifrovyText.isEmpty()) {
            throw new IllegalArgumentException("zašifrovaný text musí obsahovat aspoň jeden znak ");
        }
        // z důvodů rychlosti pro otevrenyText použiji pole znaků
        char [] otevrenyText = new char [sifrovyText.length()];
        for (int i=0; i< sifrovyText.length(); i++) {
            otevrenyText[i]=desifrujZnak(sifrovyText.charAt(i), i);
        }
        return String.valueOf(otevrenyText);
    }

    //== Soukromé metody (instancí i třídy) ===========================================
    /**
     * Metoda pro normalizaci vstupního otevřeného textu - text převedu na velká písmena, zruším diakritiku.
     * znaky, které neodpovídají anglické abecedě, mezeře a číslici nahradím prázdnou mezerou.
     */
    private String normalizuj(String otevrenyText) {
        // zrušení diakritiky - viz http://www.macroware.cz/software/unicode/
        // v Java 1.5 volání a import vypadá odlišně,
        String bezDiakritiky = Normalizer.normalize(otevrenyText, Normalizer.Form.NFD).replaceAll("[^\\p{ASCII}]","");
        // převod na velká písmena
        String velka = bezDiakritiky.toUpperCase();
        // zrušení nevhodných znaků
        return velka.replaceAll("[^A-Z]","");
    }

    /**
     * metoda pro zašifrování jednoho znaku
     */
    private char zasifrujZnak(char znak, int poradi) {
        char vysl = '\00';
        for (int i = 0; i< ABECEDA.length; i++) {
            if (ABECEDA[i]==znak) {
                int pozice = (i + posuny[poradi % posuny.length]) % ABECEDA.length;
                vysl = ABECEDA[pozice];
                break;
            }
        }
        if (vysl== '\00') {
            throw new IllegalArgumentException("neznámý znak ve vstupním poli - "+znak);
        }
        return vysl;
    }

    /**
     * Metoda pro dešifrování jednoho znaku
     */
    private char desifrujZnak(char znak, int poradi) {
        char vysl = '\00';
        for (int i = 0; i< ABECEDA.length; i++) {
            if (ABECEDA[i]==znak) {
                int pozice = ((i + ABECEDA.length) - posuny[poradi % posuny.length]) % ABECEDA.length;
                vysl = ABECEDA[pozice];
                break;
            }
        }
        if (vysl== '\00') {
            throw new IllegalArgumentException("neznámý znak v zašifrovaném textu - "+znak);
        }
        return vysl;
    }

}
